
# Utility Routines

our @Tables_to_save = qw(
    calibration
    instrument
    network
    schanloc
    sensor
    site
    sitechan
    snetsta
    stage
) ; 
our $Save_dir = "/tmp/dbb.$$" ; 
use File::Copy ; 

use strict ; 
use warnings ;

our (%Added, %NewFiles, %PrevAlternate, %ExtfileIndex ) ; 
our ( $Pf, $opt_d, $opt_v, $Configuration, @Db, $Force_add ) ;

our $Too_Old = str2epoch("1/1/1980") ;

sub empty { 
    if ( defined $_[0] && $_[0] !~ /^\s*$/ ) { 
	return 0 ; 
    } 
    return 1 ; 
}

sub perror { 
    my ($msg) = @_ ;
    my $err = errno ; 
    print STDERR "$msg: $err\n" if ( $err ) ;
}

sub save_revert { 
    makedir ( $Save_dir ) ; 
    foreach my $table ( @Tables_to_save ) {
	my @db = dblookup (@Db, 0, $table, 0, 0) ;
	my $filename = dbquery (@db, dbTABLE_FILENAME ) ; 
	my $basename = $filename ; 
	$basename =~ s".*/"" ; 
	if ( -f $filename ) { 
	    copy($filename, "$Save_dir/$basename" ) ;
	}
    }
    my $filename = dbquery (@Db, dbDATABASE_FILENAME ) . "-dbbuild" ; 
    my $basename = $filename ; 
    $basename =~ s".*/"" ; 
    if ( -f $filename ) { 
	copy($filename, "$Save_dir/$basename" ) ;
    }
}

sub revert { 
    foreach my $table ( @Tables_to_save ) {
	my @db = dblookup (@Db, 0, $table, 0, 0) ;
	my $filename = dbquery (@db, dbTABLE_FILENAME ) ; 
	my $basename = $filename ; 
	$basename =~ s".*/"" ; 
	my $copy = "$Save_dir/$basename" ; 
	if ( -f $copy ) { 
	    copy($copy, $filename ) ;
	} elsif ( -f $filename ) { 
	    unlink $filename ; 
	}
    }
    my $filename = dbquery (@Db, dbDATABASE_FILENAME ) . "-dbbuild" ; 
    my $basename = $filename ; 
    $basename =~ s".*/"" ; 
    my $copy = "$Save_dir/$basename" ; 
    if ( -f $filename ) { 
	copy($filename, "$Save_dir/$basename" ) ;
    }
    if ( -f $copy ) { 
	copy($copy, $filename ) ;
    } elsif ( -f $filename ) { 
	unlink $filename ; 
    }
    foreach my $file ( keys %NewFiles ) { 
	unlink $file ; 
	delete $NewFiles{$file} ; 
    }
}

sub cleanup_revert {
    open ( SAVED, "find $Save_dir -depth |" ) ; 
    while ( <SAVED> ) { 
	chomp ; 
	if ( -d $_ ) { 
	    rmdir $_ ; 
	} else { 
	    unlink $_ ; 
	}
    }
}

sub trim { 
    my ($w, $ref) = @_ ; 
    return if ! defined $$ref ;
    $$ref =~ s/^\s+// ; 
    $$ref =~ s/\s+$// ; 
}

sub fieldwidth { 
    my ($field) = @_ ; 
    foreach my $table ( qw(sensor instrument site sitechan) ) { 
	my @db = dblookup ( @Db, 0, $table, $field, 0 ) ; 
	if ( $db[2] >= 0 ) { 
	    return dbquery ( @db, dbFIELD_SIZE ) ;
	}
    }
    elog_die ( "Can't find field '$field' in database\n" ) ; 
}

sub nullvalue { 
    my ($field) = @_ ; 
    foreach my $table ( qw(sensor instrument site sitechan stage calibration) ) { 
	my @db = dblookup ( @Db, 0, $table, $field, 0 ) ; 
	if ( $db[2] >= 0 ) { 
	    return dbquery ( @db, dbNULL ) ;
	}
    }
    elog_die ( "Can't find field '$field' in database\n" ) ; 
}

sub save_configuration { 
    my ($filename) = @_ ; 
    my $old ; 
    if (defined $filename) { 
	open ( $old, ">&STDOUT") or elog_die "Can't dup STDOUT: $!";
	open(STDOUT, ">$filename") or elog_die "Can't open $filename: $!" ;
    }
    $Configuration->print();
    if (defined $filename) { 
	close STDOUT ; 
	open (STDOUT, ">&", $old) or elog_die "Can't restore STDOUT: $!";
    }
}

sub append_configuration { 
    my ($filename, @batch) = @_ ; 
    if ( @batch > 0 ) { 
	open(DBC, ">>$filename") or elog_die "Can't open $filename: $!" ;  
	my $err = errno ; # suprress invalid seek error when $filename isn't present
	print DBC "\n# " . $Configuration->id() . "\n" ; 
	print DBC @batch ; 
	print DBC "\n" ;
	close DBC ;
    } else { 
	my $old ;
	open ( $old, ">&STDOUT") or elog_die "Can't dup STDOUT: $!";
	open(STDOUT, ">>$filename") or elog_die "Can't open $filename: $!" ;  
	$Configuration->print();
	print STDOUT "add\n" ; 
	close STDOUT ; 
	open (STDOUT, ">&", $old) 
	    or elog_die "Can't restore original STDOUT: $!";
    }
}

sub check_site { 
    my @complaints ;
    if ( $Configuration->net =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in the network code") ;
    }
    if ( $Configuration->netname =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in the network name") ;
    }
    if ( $Configuration->station->sta =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in the short station code") ;
    }
    if ( $Configuration->time =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in the effective time for this configuration" ) ;
    } else { 
	my ($time, $msg) = is_epoch_string($Configuration->time) ; 
	if ( ! defined $time ) { 
	    push(@complaints, "  Bad Configuration time: $msg" ) ; 
	} elsif ( $time > now() + 86400 * 365 ) {
	    push (@complaints, "configuration time is more than 1 year in the future" ) ; 
	} elsif ( $time < $Too_Old ) { 
	    push (@complaints, "configuration time is too old (before 1980)" ) ; 
	}
    }
    if ( $Configuration->station->sta =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in a long descriptive name for the station") ;
    }
    if ( $Configuration->station->lat =~ /^\s*$/
	|| $Configuration->station->lon =~ /^\s*$/
	|| $Configuration->station->elev =~ /^\s*$/ ) { 
	push (@complaints, "  Please fill in the latitude, longitude and elevation of the station") ;
    } else { 
	if ( $Configuration->station->lat < -90 || $Configuration->station->lat > 90 ) { 
	    push (@complaints, "  value for latitude is invalid") ;
	}
	if ( $Configuration->station->lon < -180 || $Configuration->station->lon > 180 ) { 
	    push (@complaints, "  value for longitude is invalid") ;
	}
	if ( $Configuration->station->elev < -10 || $Configuration->station->elev > 10 ) { 
	    push (@complaints, "  value for elevation is invalid") ;
	}
    }
    if ( @complaints ) { 
	elog_complain ( "Can't add to site table\n" 
	. "More information is required about the station:\n" 
	. join("\n", @complaints) ) ;
    }
    return @complaints == 0 ; 
}

sub check_channel{ 
    my ($channel) = @_ ; 
    my @complaints = () ;
    if ($channel->loc eq "-" ) {
	push(@complaints, "loc code may not be '-'") ; 
    }
    return @complaints ;
}

sub check_configured { 
    my @complaints ;
    my $ref = $Configuration->sensors() ; 
    my $ok = 0 ; 
    foreach ( @$ref ) { 
	if ( $_->snchoice ne "" ) { 
	    $ok = 1 ; 
	}
    }
    if ( ! $ok ) { 
	push (@complaints, "  Please choose a sensor" ) ;
    }
    if ( ! defined $Configuration->datalogger  ) { 
	push (@complaints, "  Please choose the datalogger" ) ;
    }
    if ( @complaints ) { 
	elog_complain ( "Can't add yet\n" 
	    . "Can't add to sitechan, sensor, instrument and stage tables yet\n"
	    . "  need more information:\n" 
	    .  join("\n", @complaints) ) ;
    } else { 
	my $datalogger = $Configuration->datalogger ;
	if ( ! defined $datalogger->datalogger_serialnumber
	   || $datalogger->datalogger_serialnumber =~ /^\s*$/ ) {  
	    push (@complaints, "  Please fill in the datalogger serial number") ;
	}

	my $sensors = $Configuration->sensors ; 
	foreach my $sensor (@$sensors) { 
	    next if $sensor->snchoice eq "" ;
	    if ( $sensor->edepth =~ /^\s*$/ ) { 
		push (@complaints, "  Please fill in the emplacement depth for sensor #" . $sensor->index ) ;
	    }
	    if ( ! defined $sensor->sensor_serialnumber 
		  || $sensor->sensor_serialnumber =~ /^\s*$/ ) {  
		push (@complaints, "  Please fill in the sensor serial number for sensor #" . $sensor->index ) ;
	    }
	    my $streams = $sensor->samplerates ;
	    my $streamcount = 0 ; 
	    my $channelcount = 0 ;
	    foreach my $stream (@$streams) { 
		if ( specified($stream->samprate) && $stream->samprate > 0) {
		    my $channels = $stream->channels ; 
		    foreach my $channel (@$channels) { 
			if ( $channel->on ) {
			    push(@complaints, check_channel($channel)) ; 
			    $channelcount++ ; 
			}
		    }
		    $streamcount++ ;
		}
	    }
	    if ( $streamcount < 1 ) { 
		push (@complaints, "  Please choose one or more streams/sample rates for sensor #" . $sensor->index ) ;
	    }
	    if ( $channelcount < 1 ) { 
		push (@complaints, "  Please turn on one or more channels for sensor #" . $sensor->index ) ;
	    }
	}

	if ( @complaints ) { 
	    elog_complain ( "Can't add to sitechan table\n" 
		.  "More information is required about the sensor:"
		. join("\n", @complaints) ) ; 
	}

    }
    return @complaints == 0 ; 
}

# Adding to the Database routines

sub alternate {
    my ( $name, $loc, $name0, $loc0 ) = @_ ; 
    my $ref = pfget($Pf, "alternate_Name") ; 
    my %alternate_name = %$ref ; 
    $ref = pfget($Pf, "alternate_loc") ; 
    my %alternate_loc = %$ref ; 
    my $index = "alternate" . $Configuration->station->sta . $name ; 
    if ( defined $alternate_name{$name} ) { 
	$name = $alternate_name{$name} ; 
    } elsif ( defined $alternate_loc{$loc} ) { 
	$loc = $alternate_loc{$loc} ; 
    } elsif (defined $PrevAlternate{$index} ) {
	$loc = $PrevAlternate{$index}++ ;
    } else {
	$loc = $PrevAlternate{$index} = "xa" ; 
    }
    return ($name, $loc) ; 
}

sub std_channel_name {
    my ( $samprate ) = @_ ; 
    my $band ; 
    my $pf_sensor = $Configuration->selected_sensor->pf ;
    my $short_period = pfget_boolean($pf_sensor, "short_period_sensor" ) ; 
    if ( $samprate >=  80 ) { 
	if ( $short_period ) { 
	    $band = "E" ; 
	} else { 
	    $band = "H" ; 
	}
    } elsif ( $samprate >= 10 ) { 
	if ( $short_period ) { 
	    $band = "S" ; 
	} else { 
	    $band = "B" ; 
	}
    } elsif ( $samprate > 1 ) { 
	$band = "M" ; 
    } elsif ( $samprate == 1 ) { 
	$band = "L" ; 
    } elsif ( $samprate >= 0.05 ) { 
	$band = "V" ; 
    } elsif ( $samprate >= 0.005 ) { 
	$band = "U" ; 
    } else { 
	$band = "R" ; 
    } 
    my $sensor_type = pfget($pf_sensor, "sensor_type_code" ) ; 
    if ( ! defined $sensor_type || $sensor_type eq "" ) { 
	$sensor_type = "H" ; 
    }
    my $name = $band . $sensor_type ; 
    return $name ;
}

sub resolve_channel_conflicts {
    my ($this_samplerate, $name, $loc) = @_ ; 
    my $sensors = $Configuration->sensors ;
    foreach my $sensor ( @$sensors ) { 
	my $samplerates = $sensor->samplerates ; 
	foreach my $samplerate (@$samplerates) { 
	    next if $this_samplerate == $samplerate ; 
	    if ( defined $samplerate->samprate ) {
		my $channels = $samplerate->channels ; 
		my $channel = $$channels[0] ; 
		$samplerate->name = substr($channel->chan,0,2) ; 
		if ( $samplerate->name . $sensor->sensor_loc eq $name . $loc ) {
		    ($name, $loc) = alternate($name, $loc, $samplerate->name, $sensor->sensor_loc ) ;
		}
	    }
	}
    }
    return ($name, $loc) ;
}

sub choose_channel_name { 
    my ( $samplerate, $loc ) = @_ ; 
    my $name = std_channel_name ( $samplerate->samprate ) ;
    ( $name, $loc) = resolve_channel_conflicts ( $samplerate, $name, $loc ) ; 
    return ($name, $loc) ; 
}

use Digest::MD5  qw(md5 md5_hex md5_base64);

sub index_dir { 
    my ($dir) = @_ ; 
    my %index ;
    $index{directory} = $dir ; 
    elog_log "indexing $dir\n" if $opt_d ;
    foreach (<$dir/*>) {
	if ( -f $_ ) { 
	    open IN, $_ ; 
	    my @in = <IN> ; 
	    close IN ; 
	    my $digest = md5_hex(@in) ; 
	    $index{$digest} = $_ ; 
	    elog_log "md5: $_ $digest\n" if $opt_d ; 
	}

    }
    $ExtfileIndex{$dir} = \%index ;
    elog_log "ExtfileIndex{$dir} = ref = $ExtfileIndex{$dir}\n"  if $opt_d ;
    show_hash ( "index for $dir is", \%index ) if $opt_d ;
}

sub show_hash {
    my ( $msg, $ref ) = @_ ; 
    my %index = %$ref ;
    printf STDERR "\n$msg (%d entries)\n", scalar(keys %index) ; 
    foreach my $key (sort keys %index ) { 
	print STDERR "   $key => $index{$key}\n" ; 
    }
}

sub save_extfile {
    my ( $dir, $dfile, $contents, $table, @db) = @_ ; 
    $dir = "." if $dir eq "" ;
    elog_log "save_extfile: dir='$dir' dfile='$dfile' \n" if $opt_d ;
    @db = dblookup(@db, 0, $table, 0, 0) ; 
    my $tabledir = dbquery(@db, dbTABLE_DIRNAME) ;
    if ( $dir !~ m"^/" ) { 
	$dir = "$tabledir/$dir" ;
    }
    if ( ! -d $dir ) { 
	makedir ( $dir ) ; 
    }
    if ( ! defined $ExtfileIndex{$dir} ) { 
	index_dir($dir) ; 
    }
    my $ref = $ExtfileIndex{$dir} ; 
    elog_log "ExtfileIndex{$dir} = ref = $ref\n"  if $opt_d ;
    # show_hash ( "index for $dir is", $ref ) if $opt_d ;

    my $result = "" ;
    my $digest = md5_hex($contents) ; 
    elog_log "digest for $dfile is $digest\n" if $opt_d ;
    if ( defined ${$ref}{$digest} ) { 
	my $oldfile = ${$ref}{$digest} ; 
	elog_log "comparing new with $oldfile\n" if $opt_d ;
	open ( SAME, $oldfile ) ; 
	my @old = <SAME> ; 
	close SAME ;
	my $different = 0 ; 
	chomp(@old) ;
	my @file = split('\n', $contents) ;
	foreach (@file) { 
	    my $old = shift(@old) ; 
	    if ( $old ne $_ ) { 
		$different = 1 ; 
		elog_log "--> new is different: '$old' != '$_'\n" if $opt_d ; 
		last ; 
	    }
	}
	if ( $different == 0 ) {
	    $result = $oldfile ; 
	    $result =~ s".*/"" ;
	}
    }
   
    if ( $result eq "") {
	elog_log "saving new file for $dfile\n" if $opt_d ; 
	my $max = fieldwidth("dfile") - 3 ;
	if ( length($dfile) > $max ) { 
	    $dfile = substr($dfile, 0, $max) ;
	}
	my $filename = "$dir/$dfile" ;
	my $n = 1 ;
	while ( -e $filename ) { 
	    $filename = "$dir/$dfile." . $n++ ;
	}
	my $err = errno ; 
	$NewFiles{$filename} = 1; 
	open ( FILE, ">$filename" ) 
	    || elog_complain ( "Can't open $filename for writing: $!" ) ;
	print FILE $contents ; 
	close FILE ; 
	${$ref}{$digest} = $filename ; 
	$ExtfileIndex{$dir} = $ref ;
	elog_log "save_extfile: saved to '$filename': $digest \n" if $opt_d ;
	$result = $filename ; 
	$result =~ s".*/"" ;
	show_hash ( "after saving $filename, index for $dir is", $ref ) if $opt_d ;
    }
    return $result ; 
}

sub equal {
   my ($first, $second) = @_;
   no warnings;  # silence spurious -w undef complaints
   return 0 unless @$first == @$second;
   for (my $i = 0; $i < @$first; $i++) {
       return 0 if $first->[$i] ne $second->[$i];
   }
   return 1;
}


sub add2db { 
    my ($hashref, $table, @db) = @_ ;
    if ( $opt_d ) { 
	my @hash = %$hashref ;
	grep ( s"^"'", @hash ); 
	grep ( s"$"'", @hash ); 
	elog_log "adding to $table: @hash\n" ; 
    }
    my $mark = elog_mark() ;
    @db = dblookup (@db, 0, $table, 0, 0 ) ; 
    my $record ;
    if ( $db[1] < 0 ) { 
	return (-1, "can't open table $table\n") ;
    } else { 
	my $n = dbquery (@db, dbRECORD_COUNT) ;
	$db[3] = dbNULL ;
	dbget(@db, 0) ;
	$db[3] = dbSCRATCH ; 
	my @list = %$hashref ; 
	dbputv (@db, @list ) ; 
	$record = dbaddchk(@db) ; 
	if ( $record == dbINVALID() ) {
	    return (-1, "") ;
	} elsif ( $record < 0 ) { 
	    my $match = (dbINVALID) - $record - 1 ;
	    elog_log "dbaddchk returns $record, record #$match matches\n" ;
	    my @dbold = @db ; 
	    $dbold[3] = $match ; 
	    if ( $table eq "site" ) { 
		my @fields = qw(lat lon elev staname) ;
		my @old = dbgetv(@dbold, @fields) ; 
		my @new = dbgetv(@db, @fields ) ; 
		if ( ! equal ( \@old, \@new ) ) { 
		    my $offdate = dbgetv ( @dbold, "offdate") ;
		    my $null = nullvalue ( "offdate" ) ; 
		    if ( $offdate eq $null ) { 
			my $new_offdate = dbgetv(@db, "ondate" ) ; 
			my $tmp_offdate = epoch($new_offdate)-86400 ;
			$tmp_offdate = yearday($tmp_offdate) ;
			my $old_ondate = dbgetv ( @dbold, "ondate") ;
			if ( $new_offdate <= $old_ondate ) {
			    return ( -1, "Can't change old configurations." ) ; 
			}
			$new_offdate-- ; 
			dbputv ( @dbold, "offdate", $tmp_offdate ) ; 
			dbflush_indexes(@dbold) ;
			$record = dbaddchk(@db) ; 
			if ( $record < 0 ) { 
			    dbputv ( @dbold, "offdate", $null ) ; 
			    return (-1, "failed adding new record to site : ") ; 
			}
			dbputv ( @dbold, "offdate", $new_offdate ) ; 
		    } else { 
			return (-1, "Won't update old site record #$match" ) ; 
		    }
		} else { 
		    $record = $match ;
		}

	    } elsif ( $table eq "sitechan" ) { 
		my @fields = qw(hang vang edepth descrip ctype) ;
		my @old = dbgetv(@dbold, @fields) ; 
		my @new = dbgetv(@db, @fields ) ; 
		if ( ! equal ( \@old, \@new ) ) { 
		    elog_log ("stage $old[3] doesn't match: \n    @fields\n    @old\n    @new\n" ) ; 
		    my $offdate = dbgetv ( @dbold, "offdate") ;
		    my $null = nullvalue ( "offdate" ) ; 
		    if ( $offdate eq $null ) { 
			my $new_offdate = dbgetv(@db, "ondate" ) ; 
			my $tmp_offdate = epoch($new_offdate)-86400 ;
			$tmp_offdate = yearday($tmp_offdate) ;
			my $old_ondate = dbgetv ( @dbold, "ondate") ;
			if ( $new_offdate <= $old_ondate ) {
			    return ( -1, "Can't change old configurations." ) ; 
			}
			dbputv ( @dbold, "offdate", $tmp_offdate ) ; 
			dbflush_indexes(@dbold) ;
			$record = dbaddchk(@db) ; 
			if ( $record < 0 ) { 
			    dbputv ( @dbold, "offdate", $null ) ; 
			    return (-1, "failed adding new record to sitechan : ") ; 
			}
			dbputv ( @dbold, "offdate", $new_offdate ) ; 
		    } else { 
			return (-1, "Won't update old sitechan record #$match" ) ; 
		    }
		} else { 
		    $record = $match ;
		}

	    } elsif ( $table eq "sensor" ) { 
		my @fields = qw( inid chanid ) ; 
		my @old = dbgetv(@dbold, @fields) ; 
		my @new = dbgetv(@db, @fields ) ; 
		if ( ! equal ( \@old, \@new ) ) { 
		    my $endtime = dbgetv ( @dbold, "endtime") ;
		    my $null = nullvalue ( "endtime" ) ; 
		    if ( $endtime eq $null ) { 
			my $new_endtime = dbgetv(@db, "time" ) ; 
			my $old_time = dbgetv ( @dbold, "time") ;
			$new_endtime-- ; 
			dbputv ( @dbold, "endtime", $new_endtime ) ; 
			dbflush_indexes(@dbold) ;
			$record = dbaddchk(@db) ; 
			if ( $record < 0 ) { 
			    dbputv ( @dbold, "endtime", $null ) ; 
			    return (-1, "failed adding new record to sensor" ) ; 
			}
		    } else { 
			return (-1, "Won't update old sensor record #$match" ) ; 
		    }
		} else { 
		    $record = $match ;
		}

	    } elsif ( $table eq "calibration" ) { 
		my @fields = qw( insname snname dlname samprate lead calib calper fc units segtype dlsta dlchan ) ;
		my @old = dbgetv(@dbold, @fields) ; 
		my @new = dbgetv(@db, @fields ) ; 
		elog_log ( "at table eq calibration: record=$record match = $match dbold=(@dbold) db=(@db)\n" ) if $opt_d ; 
		if ( ! equal ( \@old, \@new ) ) { 
		    my $endtime = dbgetv ( @dbold, "endtime") ;
		    elog_log "got endtime=$endtime from $table : (@dbold)\n" if $opt_d ;
		    my $null = nullvalue ( "endtime" ) ; 
		    elog_log "got null endtime ='$null' : (@dbold)\n" if $opt_d ;
		    if ( $endtime eq $null ) { 
			my $new_endtime = dbgetv(@db, "time" ) ; 
			my $old_time = dbgetv ( @dbold, "time") ;
			$new_endtime-- ; 
			dbputv ( @dbold, "endtime", $new_endtime ) ; 
			dbflush_indexes(@dbold) ;
			$record = dbaddchk(@db) ; 
			if ( $record < 0 ) { 
			    dbputv ( @dbold, "endtime", $null ) ; 
			    return (-1, "failed adding new record to calibration : ") ; 
			}
		    } else { 
			return (-1, "Won't update old calibration record #$match" ) ; 
		    }
		} else { 
		    $record = $match ;
		}

	    } elsif ( $table eq "stage" ) { 
		my @fields = qw( stageid ssident gnom iunits ounits gcalib gtype izero decifac samprate leadfac dir dfile) ;
		my @old = dbgetv(@dbold, @fields) ; 
		my @new = dbgetv(@db, @fields ) ; 
		if ( ! equal ( \@old, \@new ) ) { 
		    elog_log ("stage $old[0] doesn't match: \n    @old\n    @new\n" ) ; 
		    my $endtime = dbgetv ( @dbold, "endtime") ;
		    elog_log "got endtime=$endtime from $table : (@dbold)\n" if $opt_d ;
		    my $null = nullvalue ( "endtime" ) ; 
		    elog_log "got null endtime ='$null' from $table : (@dbold)\n" if $opt_d ;
		    if ( $endtime eq $null ) { 
			my $new_endtime = dbgetv(@db, "time" ) ; 
			my $old_time = dbgetv ( @dbold, "time") ;
			$new_endtime-- ; 
			dbputv ( @dbold, "endtime", $new_endtime ) ; 
			dbflush_indexes(@dbold) ;
			elog_log("updated previously matching record, trying to add record again" ) ; 
			$record = dbaddchk(@db) ; 
			if ( $record < 0 ) { 
			    elog_log("failed to add record, restoring old endtime" ) ; 
			    dbputv ( @dbold, "endtime", $null ) ; 
			    return (-1, "failed adding new record to stage : ") ; 
			}
		    } else { 
			return (-1, "Won't update old stage record #$match" ) ; 
		    }
		} else { 
		    elog_log ("stage $old[0] matches: \n    @old\n    @new\n" ) ; 
		    $record = $match ;
		}

	    } else { 
		return (-1, "Can't add record to $table : ") ; 
	    }
	}
	my $m = dbquery (@db, dbRECORD_COUNT) ;
	if ( $table eq "calibration" && $m == $n && $Force_add ) {
	    my @dbold = @db ; 
	    $dbold[3] = $record ; 
	    my $endtime = dbgetv ( @dbold, "endtime") ;
	    elog_log "got endtime=$endtime from $table : (@dbold)\n" if $opt_d ;
	    my $null = nullvalue ( "endtime" ) ; 
	    elog_log "got null endtime ='$null' : (@dbold)\n" if $opt_d ;
	    if ( $endtime eq $null ) { 
		my $new_endtime = dbgetv(@db, "time" ) ; 
		my $old_time = dbgetv ( @dbold, "time") ;
		$new_endtime-- ; 
		dbputv ( @dbold, "endtime", $new_endtime ) ; 
		dbflush_indexes(@dbold) ;
		$record = dbaddchk(@db) ; 
		if ( $record < 0 ) { 
		    dbputv ( @dbold, "endtime", $null ) ; 
		    return (-1, "failed adding new record to calibration : ") ; 
		}
	    }
	    $m = dbquery (@db, dbRECORD_COUNT) ;
	}
	$Added{$table}++ if $m > $n ; 
    }
    elog_flush(0, $mark) ;
    return ($record, "no problems") ; 
}


sub add_network_info {
    my @db = dblookup ( @Db, 0, "network", 0, 0 ) ; 
    my $n = dbquery(@db, dbRECORD_COUNT) ;
    my $net = $Configuration->net ;
    my $retcode = 0 ;
    my %hash = (
	net => $net,
	netname => $Configuration->netname,
	) ; 
    my ($record, $msg) = add2db ( \%hash, "network", @Db ) ; 
    if ( $record < 0 ) { 
	elog_complain ( "add to network table failed: $msg" ) ; 
	$retcode = -1 ;
    }
    return $retcode ; 
}

sub add_site_info {
    my @db = dblookup ( @Db, 0, "site", 0, 0 ) ; 
    my $station = $Configuration->station ;
    $station->sta = map_seed_netsta ( $Configuration->net, $station->sta ) ; 
    my ($time, $msg) = is_epoch_string($Configuration->time) ;
    my $retcode = 0 ;
    if ( ! defined $time ) { 
	elog_log ( "  Bad Configuration time '" . $Configuration->time . "' : $msg" ) ; 
	$retcode = -1 ;
    } else { 
	my %hash = (
	    "sta", $station->sta,
	    "staname", $station->staname, 
	    "ondate", yearday($time),
	    "lat", $station->lat, 
	    "lon", $station->lon, 
	    "elev", $station->elev ) ; 
	my ($record, $msg) = add2db ( \%hash, "site", @Db ) ; 
	elog_log "add2db (site) => $record\n" if $opt_d ; 
	if ( $record < 0 ) { 
	    elog_complain ( "failed to add record to site table: record=$record\n$msg" ) ; 
	    $retcode = -1 ;
	}
    }
    return $retcode ; 
}

sub get_stage_response {
    my ( $pf, $dfile, $stagename ) = @_ ;
    my $query ;
    if ( $stagename eq "" ) { 
	$query = "response" ; 
    } else { 
	$query = $stagename . "{response}" ;
    }
    my $response = pfget($pf, $query) ; 
    if ( defined $response && $response =~ /^\*\*/ ) { 
	elog_complain ( "Can't get response from\n\t$pf.pf\n\t   $query : $response" ) ; 
    }
    elog_log "get_stage_response: pf=$pf stagename=$stagename \n" if $opt_d ; 
    $dfile .= "=$stagename" if $stagename ne "" ; 
    $dfile =~ s"/"%"g ; 
    return ( $dfile, $response ) ; 
}

sub terminate_stage { 
    my ($sta, $chan, $time, $stageid ) = @_ ; 
    my @db = dblookup ( @Db, 0, "stage", 0, 0 ) ; 
    $db[3] = dbSCRATCH ;
    my $null = nullvalue ( "endtime" ) ; 
    dbputv(@db, "sta", $sta, "chan", $chan, "time", $time, "endtime", $null, "stageid", $stageid) ; 
    my @records = dbmatches ( @db, @db, "stagehook", qw(sta chan time::endtime stageid) ) ; 
    foreach (@records) { 
	$db[3] = $_ ; 
	dbputv(@db, "endtime", $time-1 ) ; 
    }
}

sub save_stage_files { 
    my ($sta, $chan, $time, $stream, $purported_samprate, $channel) = @_ ; 
    
    # first stage is sensor
    my $stageid = 1 ;
    my $dir = pfget($Pf, "stage_dir" ) ; 
    my $sensor = $Configuration->selected_sensor ;
    my $pf_sensor = $sensor->pf ;
    my ($dfile, $response) = get_stage_response($pf_sensor, $sensor->snchoice, "") ;
    my $total_dfile = $dfile ;
    if ( $response !~ /^\s*$/ ) { 
	$dfile = save_extfile ($dir, $dfile, $response, "stage", @Db) ;
    } else { 
	elog_log("response is null for $sta:$chan ($purported_samprate) stage #$stageid") ; 
    }
    my $total_response = $response ; 
    my $gnom = $channel->component->sensor_gain ; 
    my $sensitivity = $gnom ; 
    my $sensor_units = pfget($pf_sensor, "iunits") ; 
    my %hash = ( "sta", $sta, 
		 "chan", $chan, 
		 "time", $time, 
		 "stageid", $stageid, 
		 "ssident", $sensor->sensor_serialnumber,
		 "gnom", $gnom,
		 "iunits", $sensor_units,
		 "ounits", pfget($pf_sensor, "ounits"),
		 "gtype", pfget($pf_sensor, "gtype"),
		 "dir", $dir, 
		 "dfile", $dfile ) ;  
    my ($record, $msg) = add2db ( \%hash, "stage", @Db ) ; 
    elog_log "add2db (stage) => $record\n" if $opt_d ; 
    if ( $record < 0 ) {
	elog_complain ( "add stage #1: $record : $msg" ) ; 
    }

    if ( specified($channel->component->preamp_gain) || specified($channel->component->preamp_stage)) { 
	$stageid++ ;
	if ( specified($channel->component->preamp_gain) ) { 
	    $gnom = $channel->component->preamp_gain ; 
	    $sensitivity *= $gnom 
	} else { 
	    $gnom = nullvalue("gnom") ; 
	}
	my %hash = ( "sta", $sta, 
		     "chan", $chan, 
		     "time", $time, 
		     "stageid", $stageid, 
		     "gnom", $gnom,
		     "gtype", "preamplifier",
		     ) ;  
	if ( specified($channel->component->preamp_stage)) { 
	    $total_dfile .= "+" . $channel->component->preamp_stage ;
	    my $pf_preamp = $Configuration->PreampDir . "/" .  $channel->component->preamp_stage . ".pf" ; 
	    my $dir = pfget($Pf, "stage_dir" ) ; 
	    my ($dfile, $response) = get_stage_response($pf_preamp, $channel->component->preamp_stage);
	    $dfile = save_extfile ($dir, $dfile, $response, "stage", @Db) ;
	    $hash{dir} = $dir ; 
	    $hash{dfile} = $dfile ; 
	}
	my ($record, $msg) = add2db ( \%hash, "stage", @Db ) ; 
	elog_log "add2db (stage) => $record\n" if $opt_d ; 
	if ( $record < 0 ) {
	    elog_complain ( "add stage #stageid: $record : $msg" ) ; 
	}
    }
    
    my $datalogger = $Configuration->datalogger ;
    $total_dfile .= "+" . $datalogger->dlchoice ; 

    my $pf_datalogger = $datalogger->pf ; 
    my $stages = $stream->stages ; 
    my $samprate = $purported_samprate ;
    foreach my $stagename ( @$stages ) {
	$stageid++ ;
	my $stage_samprate = pfget($pf_datalogger, "$stagename\{samprate}") ;
	if (defined $stage_samprate ) { 
	    elog_log "using stage samprate $stage_samprate instead of a/d samprate $samprate\n" if $opt_d ; 
	    $samprate = $stage_samprate ;
	}

	my ($dfile, $response) = get_stage_response($pf_datalogger, $datalogger->dlchoice, $stagename) ;
	if ( defined $response && $response !~ /^\s*$/ ) { 
	    $dir = pfget($Pf, "stage_dir" ) ; 
	    $response =~ s/\n1000\b/\n$samprate/ ; 
	    $dfile = save_extfile ($dir, $dfile, $response, "stage", @Db) ;
	    $total_response .= "#\n$response" ; 
	} else { 
	    $dir = "-" ;
	    $dfile = "-" ;
	    # elog_log "response is null for stage #$stageid\n" ; 
	}

	my $decifac = pfget($pf_datalogger, "$stagename\{decifac}") ;
	$samprate /= $decifac if defined $decifac ;
	
	$gnom = pfget($pf_datalogger, "$stagename\{gnom}") ;
	if (defined $gnom && $gnom > 0) {
	    $sensitivity *= $gnom 
	} else { 
	    $gnom = nullvalue("gnom") ; 
	}
	my $iunits = pfget($pf_datalogger, "$stagename\{iunits}") || "-" ;
	my $ounits = pfget($pf_datalogger, "$stagename\{ounits}") || "-" ;
	my $sn = pfget($pf_datalogger, "$stagename\{sn}") || "-" ;
	if ( $sn eq "datalogger" ) { 
	    $sn = $datalogger->datalogger_serialnumber ;
	}
        my $gtype = pfget($pf_datalogger, "$stagename\{gtype}") ;
	my %hash = ( "sta", $sta, 
		 "chan", $chan, 
		 "time", $time, 
		 "stageid", $stageid,
		 "gnom", $gnom,
		 "gtype", $gtype,
		 "samprate", $samprate,
		 "ssident", $sn, 
		 "iunits", $iunits,
		 "ounits", $ounits,
		 "dir", $dir, 
		 "dfile", $dfile ) ;  
	if ( $hash{gtype} =~ /fir/i ) { 
	     my $val = pfget($pf_datalogger, "$stagename\{izero}") ;
	     $hash{izero} = $val if defined $val ;
	     $val = pfget($pf_datalogger, "$stagename\{decifac}") ;
	     $hash{decifac} = $val if defined $val ;
	     $val = pfget($pf_datalogger, "$stagename\{leadfac}") ;
	     $hash{leadfac} = $val if defined $val ;
	}
	($record, $msg) = add2db ( \%hash, "stage", @Db ) ; 
	elog_log "add2db (stage) => $record\n" if $opt_d ; 
	if ( $record < 0 ) { 
	    elog_complain ( "add stage #$stageid: $record : $msg" ) ; 
	}
    }
    while ( ++$stageid < 20 ) { 
	terminate_stage($sta, $chan, $time, $stageid ) ; 
    }
    $total_dfile .= "\@$samprate" ;
    elog_log "save_stage_files: total_dfile='$total_dfile', sensitivity=$sensitivity\n" if $opt_d ; 
    return ($total_dfile, 1.0/$sensitivity, $sensor_units, $total_response) ; 
}

sub save_soh_stages { 
    my ($sta, $chan, $time, $soh, $purported_samprate) = @_ ; 
    
    # first stage is datalogger
    my $stageid = 0 ;
    my $dir = pfget($Pf, "stage_dir" ) ; 

    my $datalogger = $Configuration->datalogger ;
    my $total_dfile = $datalogger->dlchoice ;
    my $dlchoice = $datalogger->dlchoice ;
    my $total_response = "" ;
    my $pf_datalogger = $Configuration->datalogger->pf ; 
    my $gnom = pfget($pf_datalogger, "$soh\{gnom}") ;
    my $sensor_units = pfget($pf_datalogger, "$soh\{iunits}") ;
    my $sensitivity = $gnom ;
    my $samprate ;

    my ($dfile, $response) ;
    my $stages = pfget($pf_datalogger, "$soh\{stages}" ) ; 
    if (! defined $stages ) { 
	elog_complain ( "No stages for $soh\n"
	    . "The stages for channel $soh of datalogger $dlchoice are not defined" ) ;
    } else { 
	$samprate = pfget ( $pf_datalogger, "$soh\{samprate}" ) ; 
	foreach my $stagename ( split(' ', $stages ) ) { 
	    $stageid++ ;
	    ($dfile, $response) = get_stage_response($pf_datalogger, $dlchoice, $stagename) ;
	    if ( specified( $response ) ) { 
		$dir = pfget($Pf, "stage_dir" ) ; 
		$response =~ s/\n1000\b/\n$samprate/ ; 
		$dfile = save_extfile ($dir, $dfile, $response, "stage", @Db) ;
		$total_response .= "#\n$response" ; 
	    } else { 
		$dir = "-" ;
		$dfile = "-" ;
		# elog_log "response is null for stage #$stageid\n" ; 
	    }

	    my $decifac = pfget($pf_datalogger, "$stagename\{decifac}") ;
	    $samprate /= $decifac if defined $decifac ;
	    
	    $gnom = pfget($pf_datalogger, "$stagename\{gnom}") ;
	    if (defined $gnom && $gnom > 0) {
		$sensitivity *= $gnom 
	    } else { 
		$gnom = nullvalue("gnom") ; 
	    }
	    my $iunits = pfget($pf_datalogger, "$stagename\{iunits}") || "-" ;
	    my $ounits = pfget($pf_datalogger, "$stagename\{ounits}") || "-" ;
	    my $sn = pfget($pf_datalogger, "$stagename\{sn}") || "-" ;
	    if ( $sn eq "datalogger" ) { 
		$sn = $datalogger->datalogger_serialnumber ;
	    }
	    my %hash = ( "sta", $sta, 
		     "chan", $chan, 
		     "time", $time, 
		     "stageid", $stageid,
		     "gnom", $gnom,
		     "gtype", pfget($pf_datalogger, "$stagename\{gtype}"),
		     "samprate", $samprate,
		     "ssident", $sn, 
		     "iunits", $iunits,
		     "ounits", $ounits,
		     "dir", $dir, 
		     "dfile", $dfile ) ;  
	    if ( $hash{gtype} =~ /fir/i ) { 
		 my $val = pfget($pf_datalogger, "$stagename\{izero}") ;
		 $hash{izero} = $val if defined $val ;
		 $val = pfget($pf_datalogger, "$stagename\{decifac}") ;
		 $hash{decifac} = $val if defined $val ;
		 $val = pfget($pf_datalogger, "$stagename\{leadfac}") ;
		 $hash{leadfac} = $val if defined $val ;
	    }
	    my ($record, $msg) = add2db ( \%hash, "stage", @Db ) ; 
	    elog_log "add2db (stage) => $record\n" if $opt_d ; 
	    if ( $record < 0 ) { 
		elog_complain ( "add stage #$stageid: $record : $msg" ) ; 
	    }
	}
	$total_dfile .= "\@$samprate" ;
	elog_log "save_soh_stages: total_dfile='$total_dfile', sensitivity=$sensitivity\n" if $opt_d ; 
	return ($total_dfile, 1.0/$sensitivity, $sensor_units, $total_response, $samprate) ; 
    }
}


sub get_fc {
    my ($filename) = @_ ;
    my $cmd = "dbcheck_responses -fv $filename" ;
    elog_log "running '$cmd'\n" if $opt_d ;
    open ( FC, "$cmd 2>&1 |" ) || elog_complain ( "Can't run $cmd : $!" ) ;
    my @fc = <FC> ;
    close FC ; 
    my $errno = errno ; # suppress pipe error
    elog_log "dbcheck_responses returned '@fc'\n" if $opt_d ; 
    my $fc = pop(@fc) ; 
    elog_log "last line is $fc\n" if $opt_d ;
    ($fc) = split ( ' ', $fc ) ;
    elog_log "first value from last line is $fc\n" if $opt_d ;
    return $fc ; 
}


sub add_channels {
    my ($sensor) = @_ ;
    my $retcode = 0 ; 
    my %hash ;

    my $streams = $sensor->samplerates ; 
    my $sta = $Configuration->station->sta ;
    my ($time, $msg) = is_epoch_string($Configuration->time) ;
    if ( ! defined $time ) { 
	elog_log ( "  Bad Configuration time '" . $Configuration->time . "' : $msg" ) ; 
	return -1 ;
    }
    my $datalogger = $Configuration->datalogger ;
    my ($record, $samprate ) ;

  Attempt:
    foreach my $stream (@$streams) { 
	if ( specified($stream->samprate) && $stream->samprate > 0 ) { 
	    $samprate = $stream->samprate ;
	    elog_log "\n\nattempting samprate=$samprate'\n" if $opt_d ;
	    my $inid ;
	    my $channels = $stream->channels ;
	    foreach my $channel ( @$channels ) { 
		my @dbt ; 
		if ( $channel->on ) { 

		    my $loc = $channel->loc ;
		    my $chan = $channel->chan ;
		    elog_log "\n\n ** attempting channel '$chan:$loc'\n" if $opt_d ;
		    $chan = map_seed_chanloc ( $sta, $chan, $loc ) ; 
		    my @db = dblookup(@Db, 0, "stage", 0, 0) ; 
		    my $before = dbquery(@db, dbRECORD_COUNT) ; 

		    my ($dfile, $ncalib, $sensor_units, $response) = 
				save_stage_files($sta, $chan, $time, $stream, $samprate, $channel) ;

		    my $after = dbquery(@db, dbRECORD_COUNT) ; 
		    if ( $after > $before ) { 
			$Force_add = 1 ; 
		    } else { 
			$Force_add = 0 ; 
		    }

		    my  $tfile = "/tmp/.#dbbuild$$" ;
		    open (TFILE, ">$tfile" ) ;
		    print TFILE $response ; 
		    close TFILE ;
		    my $fc = get_fc ($tfile) ; 
		    unlink ( $tfile ) ;

		    %hash = ("sta", $sta,
			     "chan", $chan, 
			     "time", $time,
			     "calib", $ncalib, 
			     "fc", $fc,   
			     "units", $sensor_units
		    ) ; 
		    $hash{lead} = $channel->component->lead ; 
		    $hash{lead} = "-" if $hash{lead} eq "" ;
		    $hash{segtype} = $sensor->rsptype ;
		    $hash{dlsta} = empty($datalogger->datalogger_id) ? "-" : $datalogger->datalogger_id ;
		    $hash{dlchan} = empty($channel->id) ? "-" : $channel->id ;
		    $hash{samprate} = $samprate ; 
		    $hash{snname} =  $sensor->snchoice ;
		    $hash{dlname} =  $datalogger->dlchosen ;
		    $hash{insname} =  $sensor->sndesc . "/" . $datalogger->dldesc ;
		    ($record, $msg) = add2db ( \%hash, "calibration", @Db ) ; 
		    elog_log "add2db (calibration) => $record\n" if $opt_d ; 
		    if ( $record < 0 ) { 
			elog_complain ( "add calibration: $record : $msg" ) ; 
			last ;
		    }

		    my $dir = pfget($Pf, "response_dir" ) ; 

		    $dfile = $sensor->dfile . "+" . $datalogger->dfile . "\@$samprate" ;
		    $dfile = save_extfile ($dir, $dfile, $response, "instrument", @Db) ;
		    %hash=("insname", $sensor->sndesc . "/" . $datalogger->dldesc,
			    "ncalib", $ncalib,
			    "instype", $sensor->dfile,
			    "samprate", $samprate,
			    "band", $sensor->band,
			    "rsptype", $sensor->rsptype,
			    "dir", $dir, 
			    "dfile", $dfile ) ; 
		    ($record, $msg) = add2db ( \%hash, "instrument", @Db ) ; 
		    elog_log "add2db (instrument) => $record\n" if $opt_d ; 
		    if ( $record < 0 ) { 
			elog_complain ( "add instrument: $record : $msg" ) ; 
			last ;
		    }
		    @dbt = dblookup ( @Db, 0, "instrument", 0, 0) ; 
		    $dbt[3] = $record ;
		    $inid = dbgetv ( @dbt, "inid" ) ; 

		    %hash = ("sta", $sta,
			     "chan", $chan, 
			     "ondate", yearday($time),
			     "hang", $channel->component->hang, 
			     "vang", $channel->component->vang, 
			     "edepth", $sensor->edepth,
			     "descrip", $sensor->sensor_serialnumber . " " . $datalogger->datalogger_serialnumber,
		    ) ;
		    ($record, $msg) = add2db ( \%hash, "sitechan", @Db ) ; 
		    elog_log "add2db (sitechan) => $record\n" if $opt_d ; 
		    if ( $record < 0 ) { 
			elog_complain ( "add sitechan: $record : $msg" ) ; 
			last ;
		    }
		    @dbt = dblookup ( @Db, 0, "sitechan", 0, 0) ; 
		    $dbt[3] = $record ;
		    my $chanid = dbgetv ( @dbt, "chanid" ) ; 

		    %hash = ("sta", $sta,
			     "chan", $chan, 
			     "time", $time,
			     "jdate", yearday($time),
			     "inid", $inid, 
			     "chanid", $chanid) ;
		    ($record, $msg) = add2db ( \%hash, "sensor", @Db ) ; 
		    elog_log "add2db (sensor) => $record\n" if $opt_d ; 
		    if ( $record < 0 ) { 
			elog_complain ( "add sensor: $record : $msg" ) ; 
			elog_complain ( "sta: $sta\nchan: $chan\ntime $time\ninid $inid\nchanid $chanid") ; 
			last ;
		    }
		}
	    }
	}
    }
}

sub add_soh {
    my $retcode = 0 ; 
    my %hash ;

    my $inid ;
    my @dbt ; 
    my $loc ;
    my $stages ;
    my $stream = "soh" ;

    my $sta = $Configuration->station->sta ;
    my ($time, $msg) = is_epoch_string($Configuration->time) ; 
    if ( ! defined $time ) { 
	elog_complain("  Bad Configuration time: $msg" ) ; 
	return -1 ; 
    }
    my $datalogger = $Configuration->datalogger ;
    my ($record, $samprate) ;
    my $pf_datalogger = $Configuration->datalogger->pf ; 
    $samprate = "0" ;
    my $state_of_health = pfget($pf_datalogger, "soh") ;
    if ( defined $state_of_health ) { 

      Attempt:
	my @state_of_health = split(' ', $state_of_health ) ; 
	foreach my $soh ( @state_of_health ) { 
	    elog_log "\n\nattempting soh chan='$soh'\n" if $opt_d ;
	    $inid = -1 ; 

	    $stages = pfget($pf_datalogger, $soh) ;
	    if ( ! defined $stages ) { 
		elog_complain ( "missing SOH channel\n" 
		    . "no definition for state of health channel '$soh'" ) ; 
		next ; 
	    }

	    $loc = "" ;
	    my $chan = map_seed_chanloc ( $sta, $soh, $loc ) ; 

	    my @db = dblookup(@Db, 0, "stage", 0, 0) ; 
	    my $before = dbquery(@db, dbRECORD_COUNT) ; 

	    my ($dfile, $ncalib, $sensor_units, $response, $samprate) = 
			save_soh_stages($sta, $chan, $time, $soh, $samprate) ;

	    my $after = dbquery(@db, dbRECORD_COUNT) ; 
	    if ( $after > $before ) { 
		$Force_add = 1 ; 
	    } else { 
		$Force_add = 0 ; 
	    }

	    my $fc ;
	    if ( specified($response) ) { 
		my  $tfile = "/tmp/.#dbbuild$$" ;
		open (TFILE, ">$tfile" ) ;
		print TFILE $response ; 
		close TFILE ;
		$fc = get_fc ($tfile) ; 
		unlink ( $tfile ) ;
	    } else { 
		$fc = nullvalue("fc") ; 
	    }

	    %hash = ("sta", $sta,
		     "chan", $chan, 
		     "time", $time,
		     "calib", $ncalib, 
		     "fc", $fc,   
		     "units", $sensor_units
	    ) ; 
	    $hash{lead} = "-" ;
	    $hash{segtype} = pfget($pf_datalogger, "$soh\{rsptype}") ;
	    $hash{segtype} = "-" if ! defined $hash{segtype} ;
	    $hash{dlsta} = empty($datalogger->datalogger_id) ? "-" : $datalogger->datalogger_id ;
	    $hash{samprate} = $samprate ; 
	    $hash{snname} =  "@" . $soh ;
	    $hash{dlname} =  $datalogger->dlchosen ;
	    $hash{insname} =  $datalogger->dldesc ;
	    ($record, $msg) = add2db ( \%hash, "calibration", @Db ) ; 
	    elog_log "add2db (calibration) => $record\n" if $opt_d ; 
	    if ( $record < 0 ) { 
		elog_complain ( "add calibration: $record : $msg" ) ; 
		last ;
	    }

	    if ( $inid < 0 && specified($response) ) {
		my $dir = pfget($Pf, "response_dir" ) ; 
		$dfile = $datalogger->dfile . "+$soh" ;
		$dfile = save_extfile ($dir, $dfile, $response, "instrument", @Db) ;
		%hash=("insname", $datalogger->dldesc,
			"ncalib", $ncalib,
			"samprate", $samprate,
			"dir", $dir, 
			"dfile", $dfile ) ; 
		($record, $msg) = add2db ( \%hash, "instrument", @Db ) ; 
		elog_log "add2db (instrument) => $record\n" if $opt_d ; 
		if ( $record < 0 ) { 
		    elog_complain ( "add instrument: $msg" ) ; 
		    last ;
		}
		@dbt = dblookup ( @Db, 0, "instrument", 0, 0) ; 
		$dbt[3] = $record ;
		$inid = dbgetv ( @dbt, "inid" ) ; 
	    }

	    %hash = ("sta", $sta,
		     "chan", $chan, 
		     "ondate", yearday($time),
		     "descrip", $datalogger->datalogger_serialnumber,
	    ) ;
	    $hash{ctype} = pfget($pf_datalogger, "$soh\{ctype}") ;
	    $hash{ctype} = "-" if ! defined $hash{ctype} ;
	    ($record, $msg) = add2db ( \%hash, "sitechan", @Db ) ; 
	    elog_log "add2db (sitechan) => $record\n" if $opt_d ; 
	    if ( $record < 0 ) { 
		elog_complain ( "add sitechan: $msg" ) ; 
		last ;
	    }
	    @dbt = dblookup ( @Db, 0, "sitechan", 0, 0) ; 
	    $dbt[3] = $record ;
	    my $chanid = dbgetv ( @dbt, "chanid" ) ; 

	    %hash = ("sta", $sta,
		     "chan", $chan, 
		     "time", $time,
		     "jdate", yearday($time),
		     "inid", $inid, 
		     "chanid", $chanid) ;
	    ($record, $msg) = add2db ( \%hash, "sensor", @Db ) ; 
	    elog_log "add2db (sensor) => $record\n" if $opt_d ; 
	    if ( $record < 0 ) { 
		elog_complain ( "add sensor: $msg" ) ; 
		last ;
	    }
	}
    }
}

sub show_Added { 
    my $added = 0 ; 
    elog_log("") ;
    foreach (sort keys %Added) { 
	if ( $Added{$_} > 0 ) { 
	    $added++ ;
	    elog_log ( "Added $Added{$_} records to $_" ) ; 
	}
	$Added{$_} = 0 ; 
    }
    if ( $added == 0 ) { 
	elog_log ("No records added to database" ) ; 
    }
    elog_flush(1, 0) ;
}

sub id { 
    my ($name,$passwd,$uid,$gid,$quota,$comment,$gcos) = getpwuid($<) ;
    $gcos =~ s/,.*// ;
    my $time = now() ;
    my $date = epoch2str($time, "%A %B %d, %Y %T %Z", "" ) 
		. epoch2str($time, "(%T UTC)", "GMT" ) ; 
    chomp(my $host = `uname -n` ) ; 
    my $id = "$name\@$host  $gcos  $date" ; 
    return $id ; 
}

sub dbconfig { 
    my $dbconfig = dbquery(@Db, dbDATABASE_FILENAME) . "-dbbuild" ;
    return $dbconfig ; 
}

sub save_to_database { 
    my (@batch) = @_ ;
    my $result = ! check_site() ;
    if ( $result == 0 ) { 
	$result = &add_network_info() ; 
	$result += &add_site_info() ;
	my $configured = check_configured() ;
	if ( $result == 0 && $configured ) { 
	    my $sensors = $Configuration->sensors ; 
	    foreach my $sensor (@$sensors) { 
		next if $sensor->snchoice eq "" ;
		$Configuration->selected_sensor = $sensor ; 
		$result += &add_channels($sensor) ;
	    }
	} else { 
	    if ( $result != 0) { 
		elog_log("result from add_network + add_site_info is $result" ) ; 
	    } else { 
		elog_log("check_configured result is $configured" ) ; 
	    }
	}
	$result += &add_soh() ;
    }
    if ( $result == 0 && added() > 0 ) { 
	$Configuration->id = id() ;
	append_configuration(&dbconfig(), @batch) ;
    }
    return $result ;
}

sub clear_added { 
    foreach (sort keys %Added) { 
	$Added{$_} = 0 ; 
    }
}

sub added { 
    my $added = 0 ;
    foreach (sort keys %Added) { 
	if ( $Added{$_} > 0 ) { 
	    $added++ ;
	}
    }
    return $added ;
}

1;
